/* This file is part of the KDE project
 *
 * Copyright (C) 2010 Anderas Holzammer <andreas.holzammer@kdab.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "kwallet.h"
#include <ksharedconfig.h>
#include <kdebug.h>
#include <kdeversion.h>
#include <QtGui/QApplication>
#include <QtCore/QPointer>
#include <QtGui/QWidget>
#include <QtDBus/QtDBus>
#include <ktoolinvocation.h>

#include <assert.h>
#include <kglobal.h>
#include <kcomponentdata.h>
#include <kaboutdata.h>
#include <kconfiggroup.h>
#include <kwindowsystem.h>
#include <kstringhandler.h>

#include "kwallet_interface.h"

using namespace KWallet;

const QString Wallet::LocalWallet() {
    KConfigGroup cfg(KSharedConfig::openConfig("kwalletrc")->group("Wallet"));
    if (!cfg.readEntry("Use One Wallet", true)) {
        QString tmp = cfg.readEntry("Local Wallet", "localwallet");
        if (tmp.isEmpty()) {
            return "localwallet";
        }
        return tmp;
    }

    QString tmp = cfg.readEntry("Default Wallet", "kdewallet");
    if (tmp.isEmpty()) {
        return "kdewallet";
    }
    return tmp;
}

const QString Wallet::NetworkWallet() {
    KConfigGroup cfg(KSharedConfig::openConfig("kwalletrc")->group("Wallet"));

    QString tmp = cfg.readEntry("Default Wallet", "kdewallet");
    if (tmp.isEmpty()) {
        return "kdewallet";
    }
    return tmp;
}

const QString Wallet::PasswordFolder() {
    return "Passwords";
}

const QString Wallet::FormDataFolder() {
    return "Form Data";
}

class Wallet::WalletPrivate
{
public:
    WalletPrivate(Wallet *wallet, const QString &n)
        : q(wallet), name(n), kwconf(new KConfig("fakewallet",KConfig::SimpleConfig))
    {
        mastercfg = KConfigGroup(kwconf, "Wallet");
        curcfg = mastercfg;
    }

    ~WalletPrivate()
    {
        delete kwconf;
    }

    void walletServiceUnregistered();

    Wallet *q;
    KConfigGroup mastercfg;
    KConfigGroup curcfg;
    QString name;
    QString folder;
    KConfig *kwconf;
};


Wallet::Wallet(int handle, const QString& name)
    : QObject(0L), d(new WalletPrivate(this, name))
{
    Q_UNUSED(handle);
    // give the caller time to connect
    QMetaObject::invokeMethod(this, "walletOpened", Qt::QueuedConnection, Q_ARG(bool, true));
}


Wallet::~Wallet() {
    delete d;
}


QStringList Wallet::walletList() {
    kDebug() << "Wallet::walletList() is not implemented!";
    return QStringList();
}


void Wallet::changePassword(const QString& name, WId w) {
    Q_UNUSED(name);
    Q_UNUSED(w);
    kDebug() << "Wallet::changePassword(const QString& name, WId w) is not implemented!";
}


bool Wallet::isEnabled() {
    return true;
}


bool Wallet::isOpen(const QString& name) {
    Q_UNUSED(name);
    return true;
}


int Wallet::closeWallet(const QString& name, bool force) {
    Q_UNUSED(name);
    Q_UNUSED(force);
    return 0;
}


int Wallet::deleteWallet(const QString& name) {
    kDebug() << "Wallet::deleteWallet(const QString& name) is not implemented!";
    return -1;
}


Wallet *Wallet::openWallet(const QString& name, WId w, OpenType ot) {
    Q_UNUSED(ot);
    Q_UNUSED(w);
    Wallet *wallet = new Wallet(1, name);
    return wallet;
}


bool Wallet::disconnectApplication(const QString& wallet, const QString& app) {
    Q_UNUSED(wallet);
    Q_UNUSED(app);
    kDebug() << "Wallet::disconnectApplication(const QString& wallet, const QString& app) is not implemented!";
    return false;
}


QStringList Wallet::users(const QString& name) {
    Q_UNUSED(name);
    kDebug() << "Wallet::users(const QString& name) is not implemented!";
    return QStringList();
}


int Wallet::sync() {

    d->curcfg.sync();
    d->mastercfg.sync();
    d->kwconf->sync();

    return 0;
}


int Wallet::lockWallet() {
    kDebug() << "Wallet::lockWallet() is not implemented!";
    return -1;
}


const QString& Wallet::walletName() const {
    return d->name;
}

bool Wallet::isOpen() const {
    return true;
}


void Wallet::requestChangePassword(WId w) {
    Q_UNUSED( w );
}


void Wallet::slotWalletClosed(int handle) {
    Q_UNUSED( handle );
}


QStringList Wallet::folderList() {
    return d->mastercfg.groupList();
}


QStringList Wallet::entryList() {
    return d->curcfg.keyList();
}


bool Wallet::hasFolder(const QString& f) {
    if (d->mastercfg.isValid())
        return d->mastercfg.hasGroup(f);
    else
        return false;
}


bool Wallet::createFolder(const QString& f) {
    Q_UNUSED(f);
    return true;
}


bool Wallet::setFolder(const QString& f) {
    d->curcfg = d->mastercfg.group(f);
    d->folder = f;
    return true;
}


bool Wallet::removeFolder(const QString& f) {
    d->mastercfg.deleteGroup(f);
    return true;
}


const QString& Wallet::currentFolder() const {
    return d->folder;
}


int Wallet::readEntry(const QString& key, QByteArray& value) {
    value = d->curcfg.readEntry( key, QByteArray() );
    return 0;
}


int Wallet::readEntryList(const QString& key, QMap<QString, QByteArray>& value) {
    Q_UNUSED( key );
    Q_UNUSED( value );
    kDebug() << "Wallet::readEntryList(const QString& key, QMap<QString, QByteArray>& value) is not implemented!";
    return -1;
}


int Wallet::renameEntry(const QString& oldName, const QString& newName) {
    Q_UNUSED( oldName );
    Q_UNUSED( newName );
    kDebug() << "Wallet::renameEntry(const QString& oldName, const QString& newName) is not implemented!";
    return -1;
}


int Wallet::readMap(const QString& key, QMap<QString,QString>& value) {
    Q_UNUSED( key );
    Q_UNUSED( value );
    kDebug() << "Wallet::readMap(const QString& key, QMap<QString,QString>& value) is not implemented!";
    return -1;
}


int Wallet::readMapList(const QString& key, QMap<QString, QMap<QString, QString> >& value) {
    Q_UNUSED( key );
    Q_UNUSED( value );
    kDebug() << "Wallet::readMapList(const QString& key, QMap<QString, QMap<QString, QString> >& value) is not implemented!";
    return -1;
}


int Wallet::readPassword(const QString& key, QString& value) {
    value = KStringHandler::obscure(d->curcfg.readEntry(key, QString()));
    return 0;
}


int Wallet::readPasswordList(const QString& key, QMap<QString, QString>& value) {
    Q_UNUSED(key);
    Q_UNUSED(value);
    kDebug() << "Wallet::readPasswordList(const QString& key, QMap<QString, QString>& value) is not implemented!";
    return -1;
}


int Wallet::writeEntry(const QString& key, const QByteArray& value, EntryType entryType) {
    Q_UNUSED(key);
    Q_UNUSED(value);
    Q_UNUSED(entryType);
    kDebug() << "Wallet::writeEntry(const QString& key, const QByteArray& value, EntryType entryType) is not implemented!";
    return -1;
}


int Wallet::writeEntry(const QString& key, const QByteArray& value) {
    d->curcfg.writeEntry(key, value);
    d->curcfg.sync();
    return 0;
}


int Wallet::writeMap(const QString& key, const QMap<QString,QString>& value) {
    Q_UNUSED(key);
    Q_UNUSED(value);
    kDebug() << "Wallet::writeMap(const QString& key, const QMap<QString,QString>& value) is not implemented!";
    return 0;
}


int Wallet::writePassword(const QString& key, const QString& value) {
    d->curcfg.writeEntry(key, KStringHandler::obscure(value));
    d->curcfg.sync();
    return 0;
}


bool Wallet::hasEntry(const QString& key) {
    return d->curcfg.hasKey(key);
}


int Wallet::removeEntry(const QString& key) {
    d->curcfg.deleteEntry(key);
    d->curcfg.sync();
    return 0;
}


Wallet::EntryType Wallet::entryType(const QString& key) {
    Q_UNUSED(key);
    kDebug() << "Wallet::entryType(const QString& key) is not implemented!";
    return Wallet::Unknown;
}

bool Wallet::folderDoesNotExist(const QString& wallet, const QString& folder)
{
    Q_UNUSED(wallet);
    Q_UNUSED(folder);
    kDebug() << "Wallet::folderDoesNotExist(const QString& wallet, const QString& folder) is not implemented!";
    return false;
}


bool Wallet::keyDoesNotExist(const QString& wallet, const QString& folder, const QString& key)
{
    Q_UNUSED(wallet);
    Q_UNUSED(folder);
    Q_UNUSED(key);
    kDebug() << "Wallet::keyDoesNotExist(const QString& wallet, const QString& folder, const QString& key) is not implemented!";
    return false;
}

void Wallet::virtual_hook(int, void*) {
    //BASE::virtual_hook( id, data );
}

void Wallet::WalletPrivate::walletServiceUnregistered()
{
}

void Wallet::slotFolderUpdated(const QString& wallet, const QString& folder) {
    if (d->name == wallet) {
        emit folderUpdated(folder);
    }
}


void Wallet::slotFolderListUpdated(const QString& wallet) {
    if (d->name == wallet) {
        emit folderListUpdated();
    }
}


void Wallet::slotApplicationDisconnected(const QString& wallet, const QString& application) {
    Q_UNUSED(wallet);
    Q_UNUSED(application);
}

void Wallet::walletAsyncOpened(int tId, int handle) {
    Q_UNUSED(tId);
    Q_UNUSED(handle);
}

void Wallet::emitWalletAsyncOpenError() {
    emit walletOpened(false);
}

#include "kwallet.moc"
