/*
    Copyright (C) 2010  Kevin Funk <krf@electrostorm.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include "networking.h"
#include "networking_wince_p.h"

#include "soliddefs_p.h"

#include <QtCore/QDebug>
#include <QtCore/QMetaType>
#include <QtCore/QThread>
#include <QtCore/QTimer>

// Qt includes not part of the public API
#include <QtCore/private/qwineventnotifier_p.h>

// Windows CE specific includes
#include <Iphlpapi.h> // PIP_ADAPTER_INFO
#include <Winsock2.h> // WSAGetLastError()

SOLID_GLOBAL_STATIC(Solid::NetworkingWinCEPrivate, globalNetworkManager)

using namespace Solid;

Q_DECLARE_METATYPE(Solid::Networking::Status)

/**
 * Windows CE currently has no support for kded for various reasons
 * (additional memory consumption, unsatisfied dependencies (e.g. sycoca stuff)
 * when enabling kded support, additional dependencies to kdebase/*, etc.)
 *
 * This is why Windows CE network status support is directly implemented here.
 * Instead of calling kded over dbus when using Solid::Networking::Notifier,
 * or ...::status() we are just injecting our platform specific code here into
 * the default Networking implementation.
 *
 * This saves a lot of pain and satisfies our needs ;)
 **/
Solid::NetworkingWinCEPrivate::NetworkingWinCEPrivate()
    : m_networkingStatus(Solid::Networking::Unknown)
    , m_connectPolicy(Solid::Networking::Managed)
    , m_disconnectPolicy(Solid::Networking::Managed)
{
    qDebug() << Q_FUNC_INFO;

    qRegisterMetaType<Solid::Networking::Status>(); // for threading support

    checkStatus();

    HANDLE hAddressChange = CreateEvent(NULL, TRUE, FALSE, NULL);
    DWORD ret = NotifyAddrChange(&hAddressChange, NULL);
    if (ret != NO_ERROR) {
        if (WSAGetLastError() != WSA_IO_PENDING) {
            qWarning() << "NotifyAddrChange error:" << WSAGetLastError();
            return;
        }
    }

    QWinEventNotifier* notifier = new QWinEventNotifier(hAddressChange, this);
    connect(notifier, SIGNAL(activated(HANDLE)), SLOT(checkStatus()));
}

Solid::NetworkingWinCEPrivate::~NetworkingWinCEPrivate()
{
}

void Solid::NetworkingWinCEPrivate::checkStatus()
{
    Solid::Networking::Status bestStatus = Solid::Networking::Unknown;

    PIP_ADAPTER_INFO pAdapterInfo = createAdapterInfoIndex();
    while (pAdapterInfo) {
        Solid::Networking::Status adapterStatus = statusOfAdapter(pAdapterInfo);
        qDebug() << "Status of" << pAdapterInfo->AdapterName << "is" << adapterStatus;
        if (adapterStatus > bestStatus)
            bestStatus = adapterStatus;

        pAdapterInfo = pAdapterInfo->Next;
    }
    freeAdapterInfoIndex(pAdapterInfo);

    qDebug() << "Overall status is" << bestStatus;

    if (m_networkingStatus == bestStatus)
        return;

    m_networkingStatus = bestStatus;
    emit statusChanged(m_networkingStatus);
}

PIP_ADAPTER_INFO Solid::NetworkingWinCEPrivate::createAdapterInfoIndex()
{
    PIP_ADAPTER_INFO pAdapterInfo = 0;
    ULONG bufferLength = sizeof(IP_ADAPTER_INFO);

    pAdapterInfo = reinterpret_cast<IP_ADAPTER_INFO*>(new char[bufferLength]);

    // allocate memory
    DWORD rc = GetAdaptersInfo(pAdapterInfo, &bufferLength);
    if (rc == ERROR_BUFFER_OVERFLOW) {
        qDebug("Not enough buffer! Re-allocating...");

        delete[] reinterpret_cast<char*>(pAdapterInfo);
        pAdapterInfo = reinterpret_cast<IP_ADAPTER_INFO*>(new char[bufferLength]);
    } else if (rc == ERROR_NO_DATA) {
        qWarning("No adapter info available");
        return 0;
    } else if (rc != ERROR_SUCCESS) {
        qWarning("First call of GetAdaptersInfo() failed");
        return 0;
    }

    if (GetAdaptersInfo(pAdapterInfo, &bufferLength) == NO_ERROR) {
        return pAdapterInfo;
    } else {
        qWarning("Second call of GetAdaptersInfo() failed");
        return 0;
    }

    return 0;
}

void Solid::NetworkingWinCEPrivate::freeAdapterInfoIndex(PIP_ADAPTER_INFO adapterInfoIndex)
{
    if (!adapterInfoIndex)
        return;

    delete[] reinterpret_cast<char*>(adapterInfoIndex);
}

Solid::Networking::Status Solid::NetworkingWinCEPrivate::statusOfAdapter(PIP_ADAPTER_INFO info)
{
    if (!info)
        return Solid::Networking::Unknown;

    if (!info->CurrentIpAddress)
        return Solid::Networking::Unconnected;

    qDebug() << "IP of adapter:" << info->CurrentIpAddress->IpAddress.String;

    // validate IP address, sort out netmasks, etc.    
    if (inet_addr(info->CurrentIpAddress->IpAddress.String) == 0) 
        return Solid::Networking::Unconnected;

    return Solid::Networking::Connected;
}

Solid::Networking::Notifier::Notifier()
{
}

Solid::Networking::Status Solid::Networking::status()
{
    return static_cast<Solid::Networking::Status>( globalNetworkManager->status() );
}

Solid::Networking::Notifier *Solid::Networking::notifier()
{
    return globalNetworkManager;
}

Solid::Networking::ManagementPolicy Solid::Networking::connectPolicy()
{
    return globalNetworkManager->m_connectPolicy;
}

void Solid::Networking::setConnectPolicy( Solid::Networking::ManagementPolicy policy )
{
    globalNetworkManager->m_connectPolicy = policy;
}

Solid::Networking::ManagementPolicy Solid::Networking::disconnectPolicy()
{
    return globalNetworkManager->m_disconnectPolicy;
}

void Solid::Networking::setDisconnectPolicy( Solid::Networking::ManagementPolicy policy )
{
    globalNetworkManager->m_disconnectPolicy = policy;
}

#include "networking.moc"
#include "networking_wince_p.moc"
