/*
    Copyright (C) 2010  Kevin Funk <krf@electrostorm.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include "networkingcontrolmanager_wince.h"

#include "soliddefs_p.h"

// Windows includes, do NOT change order!
#include <objbase.h>
#include <initguid.h>
#include <connmgr.h>
#include <shellapi.h>
#include <winbase.h>
#include <windows.h>

#include <QtCore/QDebug>

#define WM_CONNMGR_STATE_CHANGED (WM_USER + 1)

using namespace Solid;

extern HINSTANCE qWinAppInst(); // from QtCore

LRESULT CALLBACK windowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
    case WM_CONNMGR_STATE_CHANGED:
        // FIXME: this could be more type-safe
        NetworkingControlManager* manager = reinterpret_cast<NetworkingControlManager*>(GetWindowLong(hWnd, GWL_USERDATA));
        QMetaObject::invokeMethod(manager, "updateStatus"); // thread-safe call of updateStatus
        break;
    }

    // not handled
    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


WinCENetworkingControlManager::WinCENetworkingControlManager(QObject* parent)
    : NetworkingControlManager(parent)
    , m_handle(NULL)
    , m_hWnd(NULL)
    , m_status(Networking::Unknown)
    , m_connectionRequested(false)
{
    initMessageWindow();

    connect(this, SIGNAL(statusChanged(Networking::Status)),
            SLOT(checkConnection()));
}

WinCENetworkingControlManager::~WinCENetworkingControlManager()
{
    destroyMessageWindow();
}

void WinCENetworkingControlManager::checkConnection()
{
    qDebug() << Q_FUNC_INFO << m_status;

    if (m_status == Networking::Disconnecting || m_status == Networking::Unconnected) {
        if (connectionRequested()) {
            qDebug() << "Warning: Manager is in disconnected state although we need a connection.";

            // TODO: Check if reconnect attempts are required on Windows CE
        }
    }
}

void WinCENetworkingControlManager::initMessageWindow()
{
    HINSTANCE hInstance = qWinAppInst();

    const wchar_t* szWindowClass = L"ServiceWindow";
    WNDCLASS wc = {
        CS_NOCLOSE,
        windowProc,
        0,
        0,
        hInstance,
        NULL,
        NULL,
        NULL,
        NULL,
        szWindowClass
    };
    RegisterClass(&wc);

    m_hWnd = CreateWindow(
        szWindowClass,
        L"ServiceProc",
        ~WS_VISIBLE,
        0, 0, 0, 0,
        NULL,
        NULL,
        hInstance,
        NULL
    );

    if (!m_hWnd) {
        qWarning() << "ERROR: Failed to create message-only window for NetworkingControlManager, no status updates possible!";
        qWarning() << "ERROR: " << GetLastError();
    }

    SetWindowLong(m_hWnd, GWL_USERDATA, (LONG)this);
}

void WinCENetworkingControlManager::destroyMessageWindow()
{
    DestroyWindow(m_hWnd);
}

void WinCENetworkingControlManager::updateStatus()
{
    DWORD status = CONNMGR_STATUS_UNKNOWN;
    ConnMgrConnectionStatus(m_handle, &status);

    Networking::Status networkingStatus = toNetworkingStatus(status);
    qDebug() << "Updating state, new status:" << networkingStatus << "- from ConnMgr status code:" << status;
    setStatus(networkingStatus);
}

void WinCENetworkingControlManager::setStatus(Networking::Status status)
{
    if (m_status == status)
        return;

    m_status = status;
    emit statusChanged(status);
}

void WinCENetworkingControlManager::establishConnection()
{
    m_connectionRequested = true;

    // check if there's a connection available
    if (m_handle != NULL) {
        qDebug() << "Establish: Already connected, not requesting new connection";
        return;
    }

    // Make sure Connection Manager is ready
    HANDLE hConnMgr = NULL;
    BOOL bAvailable = FALSE;
        hConnMgr = ConnMgrApiReadyEvent();

    if(!hConnMgr) {
        qDebug("!hConnMgr");
        return;
    }

    // Wait for 1 second to see if Connection Manager has signaled the event
    DWORD dwResult = WaitForSingleObject(hConnMgr, 1000);

    if(dwResult == WAIT_OBJECT_0)
        bAvailable = TRUE;

    // Close the event handle
    if(hConnMgr)
        CloseHandle(hConnMgr);

    // Did it connect ok?
    if (!bAvailable) {
        qDebug("!bAvailable");
        return;
    }

    // fill connection info
    CONNMGR_CONNECTIONINFO connInfo;
    ZeroMemory(&connInfo, sizeof(CONNMGR_CONNECTIONINFO));
    connInfo.cbSize = sizeof(CONNMGR_CONNECTIONINFO);
    connInfo.dwPriority = CONNMGR_PRIORITY_USERINTERACTIVE;
    connInfo.hWnd = m_hWnd;
    connInfo.uMsg = WM_CONNMGR_STATE_CHANGED;
    connInfo.dwParams = CONNMGR_PARAM_GUIDDESTNET;
    //connInfo.dwFlags = CONNMGR_FLAG_PROXY_HTTP;
    connInfo.bExclusive = FALSE;
    connInfo.bDisabled = FALSE;
    connInfo.guidDestNet = IID_DestNetInternet;

    HRESULT hr = ConnMgrEstablishConnection(&connInfo, &m_handle);
    if (hr != S_OK) {
        qDebug() << "Manager: Establishing failed, hr:" << hr;
    }
    else {
        qDebug() << "Manager: Establishing okay, hConn:" << m_handle;
    }
}

void WinCENetworkingControlManager::releaseConnection()
{
    m_connectionRequested = false;

    if (m_handle == NULL) {
        qDebug() << "Release: Not connected, doing nothing";
        return;
    }

    HRESULT hr = ConnMgrReleaseConnection(m_handle, 3);
    if (hr != S_OK) {
        qDebug() << "Manager: Release failed, hr:" << hr;
    }
    else {
        qDebug() << "Manager: Release okay, hConn:" << m_handle;
    }
    CloseHandle(m_handle);
    m_handle = NULL;
}

/**
 * Map connection manager status constants to Networking::Status
 *
 * \see http://msdn.microsoft.com/en-us/library/bb416274.aspx
 */
Networking::Status WinCENetworkingControlManager::toNetworkingStatus(DWORD connMgrStatus)
{
    switch(connMgrStatus)
    {
    case CONNMGR_STATUS_CONNECTED:
        return Networking::Connected;
    case CONNMGR_STATUS_DISCONNECTED:
        return Networking::Unconnected;
    case CONNMGR_STATUS_WAITINGCONNECTION:
        return Networking::Connecting;
    case CONNMGR_STATUS_WAITINGDISCONNECTION:
    case CONNMGR_STATUS_WAITINGCONNECTIONABORT:
        return Networking::Disconnecting;
    default:
        return Networking::Unknown;
    }
}

#include "networkingcontrolmanager.moc"
#include "networkingcontrolmanager_wince.moc"
