/*
    Copyright (C) 2010  Kevin Funk <krf@electrostorm.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include "minimalnetworkingclient.h"

#ifdef Q_OS_WINCE
#include <solid/networkingsession.h>
#endif

#include <QtCore/QMetaObject>
#include <QtCore/QMutex>
#include <QtCore/QString>
#include <QtCore/QTime>
#include <QtGui/QApplication>
#include <QtGui/QLabel>
#include <QtGui/QHBoxLayout>
#include <QtGui/QPushButton>
#include <QtGui/QScrollBar>
#include <QtGui/QTextBrowser>
#include <QtNetwork/QTcpSocket>

#include <iostream>

static MinimalNetworkingClient* s_instance = 0;
QMutex mutex;

QString toString( Solid::Networking::Status st )
{
  QString str;
  switch ( st ) {
    case Solid::Networking::Unknown:
      str = "Unknown";
      break;
    case Solid::Networking::Unconnected:
      str = "Unconnected";
      break;
    case Solid::Networking::Disconnecting:
      str = "Disconnecting";
      break;
    case Solid::Networking::Connecting:
      str = "Connecting";
      break;
    case Solid::Networking::Connected:
      str = "Connected";
      break;
  }
  return str;
}

MinimalNetworkingClient::MinimalNetworkingClient(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f)
    , m_socket(new QTcpSocket(this))
    , m_state(QAbstractSocket::UnconnectedState)
    , m_status(Solid::Networking::status())
#ifdef Q_OS_WINCE
    , m_session(new Solid::NetworkingSession(this))
#endif
{
    s_instance = this;

    // ui
    m_label = new QLabel(this);
    m_log = new QTextBrowser(this);
    m_log->setFontPointSize(5);
    QPushButton* establishButton = new QPushButton("Try to establish connection", this);
    QPushButton* releaseButton = new QPushButton("Release connection", this);
    QPushButton* closeButton = new QPushButton("Exit", this);

    QVBoxLayout* vbox = new QVBoxLayout(this);
    vbox->addWidget(m_label);
    vbox->addWidget(establishButton);
    vbox->addWidget(releaseButton);
    vbox->addWidget(m_log);
    vbox->addWidget(closeButton);

    // logic
    connect(Solid::Networking::notifier(),
            SIGNAL(statusChanged(Solid::Networking::Status)),
            SLOT(statusChanged(Solid::Networking::Status)));
    connect(m_socket,
            SIGNAL(stateChanged(QAbstractSocket::SocketState)),
            SLOT(stateChanged(QAbstractSocket::SocketState)));
    connect(establishButton, SIGNAL(clicked()), SLOT(initConnection()));
    connect(releaseButton, SIGNAL(clicked()), SLOT(releaseConnection()));
    connect(closeButton, SIGNAL(clicked()), SLOT(close()));

    showStatus();
}

MinimalNetworkingClient::~MinimalNetworkingClient()
{
    qDebug() << Q_FUNC_INFO;

    s_instance = 0;

#ifdef Q_OS_WINCE
    delete m_session;
#endif
}

void MinimalNetworkingClient::addMessage(const QString& message)
{
    QString log = m_log->document()->toPlainText();
    m_log->setPlainText(log + '\n' + message);

    // auto scroll hack
    QScrollBar* sb = m_log->verticalScrollBar();
    sb->setValue(sb->maximum());
}

void MinimalNetworkingClient::initConnection()
{
    // force close
    m_socket->close();

#ifdef Q_OS_WINCE
    m_session->establishConnection(); // call connmgr
#endif

    qDebug() << "Socket: Connecting...";
    m_socket->connectToHost("www.google.de", 80);
}

void MinimalNetworkingClient::releaseConnection()
{
    m_socket->close();

#ifdef Q_OS_WINCE
    m_session->releaseConnection(); // call connmgr
#endif
}

void MinimalNetworkingClient::stateChanged(QAbstractSocket::SocketState state)
{
    if (m_state == state)
        return;

    qDebug() << "Socket: State changed to" << state;

    m_state = state;
    showStatus();
}

void MinimalNetworkingClient::statusChanged(Solid::Networking::Status status)
{
    if (m_status == status)
        return;

    qDebug() << "Networking: Status changed to" << status;

    m_status = status;
    showStatus();
}

void MinimalNetworkingClient::showStatus()
{
    QString text = QString("Last changed: %1\nNetworking Status: %2\nSocket state: %3")
        .arg(QTime::currentTime().toString())
        .arg(toString(m_status))
        .arg(m_state);
    m_label->setText(text);
}

void messageHandler(QtMsgType type, const char* msg)
{
    mutex.lock();
    QString txt = msg;
    txt.prepend(" - ");
    txt.prepend(QTime::currentTime().toString());

    // thread-safe call of addMessage
    if (s_instance)
        QMetaObject::invokeMethod(s_instance, "addMessage", Qt::QueuedConnection, Q_ARG(QString, txt));

    std::cout << msg << std::endl;
    mutex.unlock();
}

int main(int argc, char** argv)
{
    qInstallMsgHandler(messageHandler);
    
    QApplication app(argc, argv);
    MinimalNetworkingClient widget;
    widget.show();
    return app.exec();
}

#include "minimalnetworkingclient.moc"
